# -*- coding: utf-8 -*-

"""
Library to do some cool visual stuff.
"""

from typing import List
import cv2
import numpy as np
import colorsys

from k_util import Region

__author__ = "Jakrin Juangbhanich"
__email__ = "juangbhanich.k@gmail.com"


# ======================================================================================================================
# Color Tools.
# ======================================================================================================================


def generate_colors(n,
                    saturation: float = 1.0,
                    brightness: float = 1.0,
                    hue_offset: float = 0.0,
                    hue_range: float = 1.0,
                    as_numpy: bool = False):
    """ Generate N amount of colors spread across a range on the HSV scale.
    Will return it in a numpy format. """
    hsv = [(hue_offset + hue_range * (i / n), saturation, brightness) for i in range(n)]
    colors_raw = list(map(lambda c: colorsys.hsv_to_rgb(*c), hsv))
    colors = np.array(colors_raw)
    colors *= 255

    if not as_numpy:
        colors = [(int(c[0]), int(c[1]), int(c[2])) for c in colors]

    return colors


# ======================================================================================================================
# Region Drawing Tools.
# ======================================================================================================================


def draw_regions(image: np.array,
                 regions: List[Region],
                 color=(255, 255, 255),
                 thickness: int = 2,
                 overlay: bool = False,
                 strength: float = 1.0):
    """ Draw a bounding box around each region area."""
    overlay_image = np.zeros_like(image, np.uint8) if overlay else np.copy(image)

    for i in range(len(regions)):
        r = regions[i]
        cv2.rectangle(overlay_image, (r.left, r.top), (r.right, r.bottom),
                      color=color,
                      thickness=thickness)

    if overlay:
        image = cv2.addWeighted(image, 1.0, overlay_image, strength, 0.0)
    else:
        image = cv2.addWeighted(image, 1.0 - strength, overlay_image, strength, 0.0)

    return image

  
def pixelate_region(image: np.array, regions: List[Region], blur_factor: float = 0.1):
    """ Re-sample the area within the regions to be pixelated. """
    for r in regions:
        target_image = safe_extract(image, r.left, r.right, r.top, r.bottom)
        h = target_image.shape[0]
        w = target_image.shape[1]

        pixel_h = max(1, int(h * blur_factor))
        pixel_w = max(1, int(w * blur_factor))

        target_image = cv2.resize(target_image, (pixel_w, pixel_h),
                                  interpolation=cv2.INTER_NEAREST)
        target_image = cv2.resize(target_image, (w, h), interpolation=cv2.INTER_NEAREST)
        image = safe_implant(image, target_image, r.left, r.right, r.top, r.bottom)

    return image


def draw_region_mask(image: np.array, regions: List[Region], strength: float = 1.0):
    """ Apply a mask to the areas covered by the regions. """

    image = image.astype(np.float)
    pos_mask = np.zeros((image.shape[0], image.shape[1]), dtype=np.bool)
    neg_mask = np.ones((image.shape[0], image.shape[1]), dtype=np.bool)

    for r in regions:
        safe_left, safe_right, _, _ = _get_safe_bounds(r.left, r.right, image.shape[1])
        safe_top, safe_bottom, _, _ = _get_safe_bounds(r.top, r.bottom, image.shape[0])

        pos_mask[safe_top:safe_bottom, safe_left:safe_right] = True
        neg_mask[safe_top:safe_bottom, safe_left:safe_right] = False

    fade_factor = 1.0 - (0.7 * strength)
    image[neg_mask] *= fade_factor
    image = image.astype(np.uint8)
    return image


# ======================================================================================================================
# Progress (or custom) Bars.
# ======================================================================================================================


def draw_bar(image, progress: float, x: int, y: int, width: int, height: int,
             frame_color=(0, 0, 0), bar_color=(0, 150, 255)):
    """ Draw a rectangular bar, with the specified progress value filled out. """
    draw_bar_segment(image, 0.0, 1.0, x, y, width, height, frame_color)  # Frame
    draw_bar_segment(image, 0.0, progress, x, y, width, height, bar_color)  # Bar


def draw_bar_segment(image, p_start: float, p_end: float, x: int, y: int, width: int, height: int, color=(0, 150, 255)):
    """ Draw a segment of a bar, for example if we just wanted a start-end section.
    Starting point is the top left. """

    # Draw the bar.
    p_width = max(0.05, p_end - p_start)
    p_width = int(width * p_width)
    p_x = int(x + p_start * width)
    cv2.rectangle(image, (p_x, y), (p_x + p_width, y + height), color, thickness=-1)


# ===================================================================================================
# 2D Image Slice Helpers.
# ===================================================================================================


def _get_safe_bounds(near: int, far: int, max_bound: int) -> (int, int, int, int):
    """ Finds the near/far bounds (such as left, right) for a certain max bound (width, etc). """
    safe_near = max(0, near)
    safe_far = min(max_bound, far)
    near_excess = safe_near - near
    far_excess = far - safe_far
    return safe_near, safe_far, near_excess, far_excess


def safe_extract(image: np.array, left: int, right: int, top: int, bottom: int):
    """ Extract the specified area from the image, padding the over-cropped areas with black.
    Assumes a np.array (CV2 image) input format. """

    safe_left, safe_right, left_excess, right_excess = _get_safe_bounds(left, right, image.shape[1])
    safe_top, safe_bottom, top_excess, bottom_excess = _get_safe_bounds(top, bottom, image.shape[0])

    # Extract the image.
    extracted_image = image[safe_top:safe_bottom, safe_left:safe_right]

    # Get the extraction area.
    output_h = bottom - top
    output_w = right - left

    safe_h = safe_bottom - safe_top
    safe_w = safe_right - safe_left

    # Fill the excess area with black.
    filler = np.zeros((output_h, output_w, 3), dtype=np.uint8)
    insert_bottom = top_excess + safe_h
    insert_right = left_excess + safe_w
    filler[top_excess:insert_bottom, left_excess:insert_right] = extracted_image
    return filler


def safe_implant(dst_image: np.array, src_image: np.array, left: int, right: int, top: int, bottom: int):
    """ Plant the area from the src image into the dst image. """

    # Get the extraction area.
    safe_left, safe_right, left_excess, right_excess = _get_safe_bounds(left, right, dst_image.shape[1])
    safe_top, safe_bottom, top_excess, bottom_excess = _get_safe_bounds(top, bottom, dst_image.shape[0])

    s_bottom = top_excess + (safe_bottom - safe_top)
    s_right = left_excess + (safe_right - safe_left)

    if safe_bottom - safe_top > 0 and safe_right - safe_left > 0:
        # Prevent import/slicing into 0 size matrices.
        dst_image[safe_top:safe_bottom, safe_left:safe_right] = src_image[top_excess:s_bottom, left_excess:s_right]

    return dst_image


def safe_extract_with_region(image: np.array, region: Region) -> np.array:
    """ Extract the image area specified by the region. """
    return safe_extract(image, region.left, region.right, region.top, region.bottom)


def safe_implant_with_region(dst_image: np.array, src_image: np.array, region: Region) -> np.array:
    """ Plant the area from the src image into the dst image. """
    return safe_implant(dst_image, src_image, region.left, region.right, region.top, region.bottom)



# -*- coding: utf-8 -*-

"""
Draw good looking text to Cv2/Numpy images. Cv2's default text renderer is a bit ugly. I can't
specify the font, and not all sizes look great. Here I use PIL with Cv2 to get ttf fonts into the jam,
and also include support for the FontAwesome icon fonts.
"""

import os
from typing import Dict
import cv2
import numpy as np
from PIL import Image, ImageFont, ImageDraw
from PIL.ImageFont import FreeTypeFont
from k_util import Region

__author__ = "Jakrin Juangbhanich"
__email__ = "juangbhanich.k@gmail.com"

# ======================================================================================================================
# Constants.
# ======================================================================================================================


ALIGN_LEFT = -1
ALIGN_CENTER = 0
ALIGN_RIGHT = 1
ALIGN_TOP = -1
ALIGN_BOTTOM = 1

FONT_DEFAULT = "DEFAULT"
FONT_ICON = "ICON"

DEFAULT_PAD: int = 8


# ======================================================================================================================
# This will manage the PIL TrueType fonts.
# ======================================================================================================================


class TextManager:

    # TODO: I might want to create a way to custom initialize this Singleton.
    # Then I can maybe specify things like custom fonts, etc.

    INSTANCE = None

    def __init__(self):

        self.base_path = os.path.join(os.path.dirname(__file__), "fonts")
        self.fonts_by_size: Dict[int, dict] = {}

        self.font_path_map: Dict[str, str] = {
            FONT_DEFAULT: "OpenSans-Semibold.ttf",
            FONT_ICON: "fa-solid-900.ttf"
        }

        self.font_divisor_map: Dict[str, float] = {
            # Some fonts need a bit more y-offset to look vertically aligned.
            FONT_DEFAULT: 2.0,
            FONT_ICON: 2.0
        }

    @staticmethod
    def instance() -> 'TextManager':
        if TextManager.INSTANCE is None:
            TextManager.INSTANCE = TextManager()
        return TextManager.INSTANCE

    @staticmethod
    def get_font(font_type: str = FONT_DEFAULT, font_size_id: int = 18) -> FreeTypeFont:
        text_manager = TextManager.instance()
        text_manager._load_font(font_type, font_size_id)
        return text_manager.fonts_by_size[font_size_id][font_type]

    @staticmethod
    def get_font_divisor(font_type: str = FONT_DEFAULT) -> float:
        text_manager = TextManager.instance()
        return text_manager.font_divisor_map[font_type]

    def _load_font(self, font_type: str = FONT_DEFAULT, size: int = 18) -> None:
        if size not in self.fonts_by_size:
            self.fonts_by_size[size] = {}

        font_dict = self.fonts_by_size[size]
        if font_type not in font_dict:
            font_path = os.path.join(self.base_path, self.font_path_map[font_type])
            font_dict[font_type] = ImageFont.truetype(font_path, size)


# ===================================================================================================
# Low Level Functions.
# ===================================================================================================


def raw_text(
        image: np.array,
        text: str,
        x: int,
        y: int,
        font_type: str = FONT_DEFAULT,
        font_size: int = 18,
        color=(255, 255, 255)
):
    """ Draw the specified text into the image at the point of the region. """

    # Convert image from CV2 to PIL.
    pil_image, pil_draw = _cv2_to_pil(image)

    # Prepare the font.
    font = TextManager.get_font(font_type=font_type, font_size_id=font_size)

    # Prepare the anchors.
    anchor_x = x
    anchor_y = y

    # Draw the text straight into the region.
    pil_draw.text((anchor_x, anchor_y), text, font=font, fill=(color[2], color[1], color[0]))

    # Convert image from PIL back to CV2.
    final_image = _pil_to_cv2(pil_image)
    return final_image


def raw_icon(
        image: np.array,
        text: str,
        x: int,
        y: int,
        font_size: int = 18,
        color=(255, 255, 255)
):
    """ Write a raw icon to the specified location. """
    return raw_text(image, text, x, y, font_type=FONT_ICON, font_size=font_size, color=color)


def write_icon(
        image: np.array,
        text: str,
        x: int,
        y: int,
        font_size: int = 18,
        color=(255, 255, 255)
):
    """ Write a centered icon at the specific location. """
    i_width, i_height = get_text_size(text, FONT_ICON, font_size)
    x -= i_width // 2
    y -= i_height // 2
    return raw_text(image, text, x, y, font_type=FONT_ICON, font_size=font_size, color=color)


def write_into_region(
        image: np.array,
        text: str,
        region: Region,
        icon: str = None,  # Inline icon to render.
        pad: int = DEFAULT_PAD,
        h_align: int = ALIGN_CENTER,
        font_type: str = FONT_DEFAULT,
        font_size: int = 18,
        color=(255, 255, 255),
        bg_color=None,
        bg_opacity=1.0,
        show_region_outline: bool = False,
        fixed_width: bool = False,  # If the width of the region was fixed from outside. Used to align the icon.
        overlay: bool = False
):
    """ The text will be written into this specified region.
    The y position will be centered. The x position will depend on the align type. """

    # Draw the BG into position.
    image = _fill_region(image, region, bg_color=bg_color, bg_opacity=bg_opacity)

    # If it is overlay, recursive call.
    if overlay:
        sub_sample = np.zeros((image.shape[0], image.shape[1], 3), dtype=np.uint8)
        overlay_image = \
            write_into_region(image=sub_sample, text=text, region=region, icon=icon, pad=pad, h_align=h_align,
                              font_type=font_type, font_size=font_size, color=color, bg_color=(0, 0, 0),
                              bg_opacity=1.0, show_region_outline=False, fixed_width=fixed_width)
        final_image = cv2.addWeighted(image, 1, overlay_image, 1, 0)
        return final_image

    # Use the region to find the position.
    t_width, t_height, i_width, i_height, b_width, b_height = \
        _get_text_and_icon_size(text, icon, pad, font_type, font_size)

    # Get the fixed height so Y axis isn't distorted.
    f_height = t_height if font_type == FONT_ICON else _get_text_fixed_height(font_type, font_size)

    # Default case is central align.
    ix = region.x - b_width // 2
    iy = region.y - i_height // 2
    ty = region.y - int(f_height / TextManager.instance().get_font_divisor(font_type))
    tx = region.x - t_width // 2

    # Left align case.
    if h_align == ALIGN_LEFT or fixed_width:
        ix = region.left if icon is None else region.left + pad

    # Align the text to the icon.
    if h_align == ALIGN_LEFT or (not fixed_width and icon is not None):
        tx = ix + i_width + pad

    # Right align case.
    if h_align == ALIGN_RIGHT:
        ix = region.right - b_width - pad
        tx = ix + i_width + pad if icon is not None else ix

    # Write the text.
    image = raw_text(image=image, text=text, x=tx, y=ty, font_type=font_type, font_size=font_size,
                     color=color)

    if icon is not None:
        image = raw_icon(image=image, text=icon, x=ix, y=iy, font_size=font_size, color=color)

    # Show an outline around the region.
    if show_region_outline:
        cv2.rectangle(image, (region.left, region.top), (region.right, region.bottom), color=(0, 255, 0), thickness=1)

    return image


# ======================================================================================================================
# High Level Functions.
# ======================================================================================================================


def center_at_position(image: np.array, text: str, x: int = 0, y: int = 0, width: int = None, height: int = None,
                       icon: str = None, pad: int = DEFAULT_PAD, font_type: str = FONT_DEFAULT, font_size: int = 18,
                       color=(255, 255, 255), bg_color=(0, 0, 0), bg_opacity=0.5, overlay: bool=False):
    return write_at_position(image=image, text=text, x=x, y=y, width=width, height=height, icon=icon, pad=pad,
                             h_align=ALIGN_CENTER, font_type=font_type, font_size=font_size, color=color,
                             bg_color=bg_color, bg_opacity=bg_opacity, overlay=overlay)


def left_at_position(image: np.array, text: str, x: int = 0, y: int = 0, width: int = None, height: int = None,
                     icon: str = None, pad: int = DEFAULT_PAD, font_type: str = FONT_DEFAULT, font_size: int = 18,
                     color=(255, 255, 255), bg_color=(0, 0, 0), bg_opacity=0.5, overlay: bool=False):
    return write_at_position(image=image, text=text, x=x, y=y, width=width, height=height, icon=icon, pad=pad,
                             h_align=ALIGN_LEFT, font_type=font_type, font_size=font_size, color=color,
                             bg_color=bg_color, bg_opacity=bg_opacity, overlay=overlay)


def write_at_position(
        image: np.array,
        text: str,
        x: int = 0,
        y: int = 0,
        width: int = None,
        height: int = None,
        icon: str = None,
        pad: int = DEFAULT_PAD,
        h_align: int = ALIGN_CENTER,
        font_type: str = FONT_DEFAULT,
        font_size: int = 18,
        color=(255, 255, 255),
        bg_color=None,
        bg_opacity=1.0,
        overlay: bool=False
):
    """ Create a center-locked region, with the specified width and height. """
    region: Region = Region(0, 10, 0, 10)
    is_fixed_width = False

    # Get the meta-data of the text boxes.
    t_width, t_height, i_width, i_height, b_width, b_height = \
        _get_text_and_icon_size(text, icon, pad, font_type, font_size)

    # Finalize the region width and height.
    if width is not None:
        region.width = width
        is_fixed_width = True
    else:
        region.width = b_width
    region.height = height if height is not None else b_height

    # Expand for the padding.
    region.width += pad * 2
    region.height += pad * 2
    region.x = x
    region.y = y

    if h_align == ALIGN_LEFT:
        region.right = x + region.width
        region.left = x

    # Draw the text.
    image = write_into_region(image=image, text=text, region=region, icon=icon, pad=pad, h_align=h_align,
                              font_type=font_type, font_size=font_size, color=color, bg_color=bg_color,
                              bg_opacity=bg_opacity, show_region_outline=False, fixed_width=is_fixed_width,
                              overlay=overlay)

    return image


def write_anchored(image: np.array, text: str, h_anchor: int = ALIGN_CENTER, v_anchor: int = ALIGN_CENTER,
                   icon: str = None, pad: int = DEFAULT_PAD, font_type: str = FONT_DEFAULT, font_size: int = 18,
                   color=(255, 255, 255), bg_color=(0, 0, 0), bg_opacity=1.0, overlay: bool = False):
    """ Find the anchored region and create a text box there. """

    image_width = image.shape[1]
    image_height = image.shape[0]

    x = image_width // 2
    y = image_height // 2

    t_width, t_height, i_width, i_height, b_width, b_height = \
        _get_text_and_icon_size(text, icon, pad, font_type, font_size)

    if h_anchor == ALIGN_LEFT:
        x = pad

    if h_anchor == ALIGN_RIGHT:
        x = image_width - (2 * pad + b_width // 2)

    if v_anchor == ALIGN_TOP:
        y = 2 * pad + b_height // 2

    if v_anchor == ALIGN_BOTTOM:
        y = image_height - (2 * pad + b_height // 2)

    return write_at_position(image=image, text=text, x=x, y=y, width=None, height=None, icon=icon, pad=pad,
                             h_align=h_anchor, font_type=font_type, font_size=font_size, color=color,
                             bg_color=bg_color, bg_opacity=bg_opacity, overlay=overlay)


def label_region(image: np.array, text: str, region: Region, icon: str = None, pad: int = 5, gap: int = 5,
                 font_type: str = FONT_DEFAULT, font_size: int = 14, show_at_bottom: bool=False,
                 color=(255, 255, 255), bg_color=(0, 0, 0), bg_opacity: float = 0.7, overlay: bool = False,
                 inside: bool=False):

    # Find the text, icon and box positions.
    t_width, t_height, i_width, i_height, b_width, b_height = \
        _get_text_and_icon_size(text, icon, pad, font_type, font_size)

    draw_region: Region = region.clone()
    draw_region.height = b_height + pad * 2

    if inside:
        draw_region.width -= gap * 2

    if draw_region.width < b_width + pad * 2:
        draw_region.width = b_width + pad * 2

    draw_direction = -1 if inside else 1
    draw_anchor = draw_region.height // 2 + gap

    # Find the place to draw, relative to the region label.
    if not show_at_bottom:
        y = region.top - (draw_direction * draw_anchor)
    else:
        y = region.bottom + (draw_direction * draw_anchor)

    draw_region.y = y

    return write_into_region(image=image, text=text, region=draw_region, icon=icon, pad=pad, h_align=ALIGN_CENTER,
                             font_type=font_type, font_size=font_size, color=color, bg_color=bg_color,
                             bg_opacity=bg_opacity, show_region_outline=False, fixed_width=True, overlay=overlay)


# ===================================================================================================
# Support Functions.
# ===================================================================================================


def get_text_size(text: str, font_type: str = FONT_DEFAULT, font_size: int = 16):
    """ Returns the width and height for this text, font and size. """
    font = TextManager.get_font(font_type=font_type, font_size_id=font_size)
    return font.getsize(text)


def _get_text_and_icon_size(text: str, icon: str, pad: int = 0, font_type: str = FONT_DEFAULT, font_size: int = 16):
    """ Get the meta-data of the text boxes. """
    t_width, t_height = get_text_size(text, font_type, font_size)
    i_width, i_height = (0, 0) if icon is None else get_text_size(icon, FONT_ICON, font_size)
    b_width, b_height = (t_width, t_height) if icon is None else (t_width + i_width + pad, max(t_height, i_height))
    return t_width, t_height, i_width, i_height, b_width, b_height


def _get_text_fixed_height(font_type: str = FONT_DEFAULT, font_size: int = 16):
    """ Get the fixed height for this font. """
    # Use a string that maximizes the font height.
    _, t_height = get_text_size("Ap", font_type, font_size)
    return t_height


def _fill_region(image: np.array, region: Region, bg_color=None, bg_opacity: float = 1.0):
    """ Fill the region in this image with a color and opacity. """

    # No color or opacity is clear, do nothing.
    if bg_color is None or bg_opacity <= 0.0:
        return image

    # Opacity 1, just draw it onto the image.
    if bg_opacity >= 1.0:
        cv2.rectangle(image, (region.left, region.top), (region.right, region.bottom), color=bg_color, thickness=-1)
        return image

    # Opacity is semi-clear. Draw it in a different image and overlay it on.
    overlay_image = np.copy(image)
    cv2.rectangle(overlay_image, (region.left, region.top), (region.right, region.bottom), color=bg_color, thickness=-1)
    return cv2.addWeighted(image, 1.0 - bg_opacity, overlay_image, bg_opacity, 0.0)


def _cv2_to_pil(image: np.array) -> (Image, ImageDraw):
    """ Convert from a PIL ImageDraw to Cv2 Numpy. """
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    pil_image = Image.fromarray(image)
    image_draw = ImageDraw.Draw(pil_image)
    return pil_image, image_draw


def _pil_to_cv2(image: Image) -> np.array:
    """ Convert from a PIL ImageDraw back to Cv2 Numpy. """
    out_image = np.array(image)
    out_image = cv2.cvtColor(out_image, cv2.COLOR_RGB2BGR)
    return out_image

